import skypager from '../index'
import Asset from './asset'
import docblock from 'docblock-parser'
import { parser, transformer, compiler } from './parsers/babel'
import indexer from './script/indexer'

import { noConflict, defaults, lazy, hidden } from '../util'

import get from 'lodash/get'
import result from 'lodash/result'
import mapValues from 'lodash/mapValues'
import isString from 'lodash/isString'
import isError from 'lodash/isError'
import isArray from 'lodash/isArray'
import attempt from 'lodash/attempt'

const EXTENSIONS = ['js', 'jsx', 'cjsx', 'coffee', 'es6']
const GLOB = '**/*.{' + EXTENSIONS.join(',') + '}'

import { join } from 'path'

/**
 * The Script Asset is a wrapper around es6 javascript files.
 *
 * Scripts are parsed using the babylon parser and compiled
 * with babel's transformer using the skypager preset.
 *
 * Script builds a metadata structure combining the information
 * we pull from the parser's AST, by parsing the docblocks,
 * as well as things such as where the asset lives
 * within the folder structure.
 *
 * @name Script
 * @extends Asset
 *
*/
export class Script extends Asset {
  static EXTENSIONS = EXTENSIONS;
  static GLOB = GLOB;

  constructor(uri, options = {}) {
    super(uri, options)

    this.transformer = (...args) => transformer(this.raw, this)
    this.indexer = (...args) => indexer(this.parsed, this)

    this.lazy('compiled', () => compiler(this.readSync().raw, this))
  }

  static decorateCollection(collection, options = {}) {
    defineProperty(collection, 'compiledPath', {
      enumerable: false,
      configurable: true,
      get: function () {
         return collection.project.paths.join('lib')
      }
    })
  }

  get compiledPath () {
    return this.paths.compiled || join(this.collection.compiledPath, this.paths.relative)
  }

  get requiredFromCompiledSource () {
    try {
      return require(this.compiledPath)
    } catch(error) {
      return false
    }
  }

  get requiredFromSource () {
    try {
      return require(this.paths.absolute)
    } catch(error) {
      return false
    }
  }

  get requireStatus() {
    this.readSync()

    let abs = this.paths.absolute
    let compiled = this.compiledPath

    return {
      raw: !isError(attempt(() => require(abs))),
      compiled: !isError(attempt(() => require(compiled))) && !this.project.readFileSync(compiled).toString().match(/Error while compiling/)
    }
  }

  compileToDisk (outputPath) {
    outputPath = outputPath || this.compiledPath

    return this.project.ensureFile(outputPath)
      .then(() => this.project.outputFile(outputPath, this.compiled))
      .then(() => outputPath)
  }

  compileToDiskSync (outputPath) {
    outputPath = outputPath || this.compiledPath

    this.project.ensureFileSync(outputPath)
    this.project.outputFileSync(outputPath, this.compiled)

    return this
  }

  parser(...args) {
    try {
      return parser(this.raw, this)
    } catch(error) {
      console.log('Parser Error', error)
    }
  }

  getData () {
    if ((!this.raw || this.raw.length === 0)) {
      this.runImporter('disk', {asset: this, sync: true})
    }

    return this.metadata || {}
  }

  get parserOptions() {
    return ParserOptions
  }

  get transformerOptions() {
    return TransformerOptions
  }

  get exports() {
    return get(this.modulesMetadata, 'exports.exported', [])
  }

  get imports() {
    return get(this.modulesMetadata, 'imports', []).map(i => i.source)
  }

  get classPropertyNames() {
    if(this.indexed){
      return keys(this.indexes._classProperties)
    }
  }

  getClassPropertyValue(name) {
    if(this.indexed && this.indexes._classProperties[name]) {
      let prop = this.indexes._classProperties[name]

      return prop.value && prop.value.type === 'StringLiteral'
        ? prop.value.extra.rawValue
        : prop.value
    }
  }

  get reactDisplayName() {
    if (this.isReactComponent) {
      return this.getClassPropertyValue('displayName')
    }
  }

  get hasReactPropTypes() {
     return (this.isReactComponent
      && this.classPropertyNames.indexOf('propTypes') >= 0
      && this.getClassPropertyValue('propTypes').properties
     )
  }

  get reactPropTypesDeclaration () {
    if (this.hasReactPropTypes) {
      return this.getClassPropertyValue('propTypes')
    }
  }

  get reactPropDocs() {
    if(this.hasReactPropTypes) {
      return this.reactPropTypesDeclaration.properties
        .map(prop => prop.leadingComments)
    }
  }

  get reactPropNames() {
    if(this.hasReactPropTypes) {
      return this.getClassPropertyValue('propTypes').properties.map(node => {
        return node.key && node.key.name
      })
    }
  }

  get isReactComponent() {
    if (this.defaultExportType === 'ClassDeclaration') {
      let superClass = result(this, 'defaultExportNode.declaration.superClass.name')

      if (superClass && superClass.match(/Component$/)) {
         return true
      }
    }
  }

  /**
   * @todo replace with nodes vm module
   */
  evalWithHelpers(helpers = {}) {
    let project = this.project
    let script = this

    helpers = defaults(helpers, {
      asset: script,
      project,
      script,
      skypager
    })

    let fn = noConflict(async function() {
      try {
        eval(script.compiled)
      } catch(error) {
        project.error('error in evaluating script', error)
      }
    }, helpers)

    fn()
  }

  get exportedClassNames () {
    if(this.indexed) {
      let exports = this.indexes._exportNodes

      return exports.filter(
        node => get(node, 'declaration.type') === 'ClassDeclaration'
      ).filter(
        node => get(node, 'declaration.id.type') === 'Identifier'
      ).map(
        node => get(node, 'declaration.id.name')
      )
    }
  }

  get importsLocalModules () {
    if (this.indexed) {
      return this.importNodes.filter(
        n => n.source && n.source.value.match(/^\./)
      ).map(n => n.source.value)
    }
  }

  get importsModuleNames () {
    if (this.indexed && this.importNodes) {
      return this.importNodes.map(n => n.source && n.source.value)
    }
  }

  get importNodes () {
    if(this.indexed) {
     return this.indexes._importNodes
    }
  }

  get exportNodes () {
    if(this.indexed) {
     return this.indexes._exportNodes
    }
  }

  get hasDefaultDocblock() {
    return this.defaultExportNode && this.defaultExportNode.leadingComments
  }

  get defaultDocblock() {
    if (this.indexed && this.hasDefaultDocblock) {
      let leadingComments = this.defaultExportNode && this.defaultExportNode.leadingComments

      let blockContent = leadingComments.map(c => c.value)
        .filter(v => v && v.length > 0)
        .join("\n")

      return docblock.parse(blockContent)
    }
  }

  get defaultExportNode () {
    if(this.indexed) {
      return this.exportNodes.find(d => d.type === 'ExportNamedDeclaration')
    }
  }

  get defaultExportName () {
    return result(this, 'defaultExportNode.declaration.id.name')
  }

  get defaultExportType () {
    return result(this, 'defaultExportNode.declaration.type')
  }

  get modulesMetadata() {
    if(this.transformerMetadata) {
      return this.transformerMetadata
    } else {
      return {
        imports:[],
        exports:{
          exported:[]
        }
      }
    }
  }

  get transformerMetadata() {
    return get(this.transformed, 'metadata.modules')
  }

  /**
   * @alias modulesMetadata
   */
  get moduleMetadata() {
    return this.modulesMetadata
  }

  get programBody() {
    return get(this.parsed, 'program.body')
  }

  get frontmatter() {
     return this.metadata
  }

  get docDescription() {
    return this.get('defaultDocblock.text', '')
  }

  get tags() {
    let tags = this.get('defaultDocblock.tags', {})

    return mapValues(tags || {}, (v,k) => {
      // coerce pluralized tags into arrays
      if (isString(v) && k.match(/s$/)) {
        return [v.trim()]
      }

      if (isString(v)) {
        return v.trim()
      }

      if (isArray(v)) {
         return v.map(val => isString(val) ? val.trim() : val)
      }

      return v
    })
  }

  get metadata() {
    return {
      ...( this.tags ),
      docDescription: this.docDescription,
      exports: this.exports,
      imports: this.imports,
      depth: this.depth,
      componentInfo: this.componentInfo,
      asset: this.assetMetadata
    }
  }

  get componentInfo() {
     return this.isReactComponent
       ? { displayName: this.reactDisplayName, propNames: this.reactPropNames }
       : {}
  }

  generateId() {
     let base = `${this.paths.relative.replace(this.extension, '')}`

     if (base.match(/\/index$/)) {
       base = base.replace(/\/index$/,'')
     }

     return base
  }
}

export default Script

export const ParserOptions = {
  sourceType: 'module',
  plugins:[
    'jsx',
    'asyncFunctions',
    'classConstructorCall',
    'doExpressions',
    'trailingFunctionCommas',
    'objectRestSpread',
    'decorators',
    'classProperties',
    'exportExtensions',
    'asyncGenerators',
    'functionBind',
    'functionSent'
  ]
}

export const TransformerOptions = {
  presets:[
    (global.$skypager && global.$skypager.babelPresetSkypager) || 'babel-preset-skypager'
  ]
}

const { defineProperty, keys } = Object
