import './index'

import Registry from './registry'
import Collection from './collection'
import VirtualMachine from './vm'
import resolver from './resolver'

import * as Helpers from './helpers'
import * as Assets from './assets'
import * as util from './util'

import logger from './logger'
import vault from './vault'
import invariant from 'invariant'

import pathExists from 'path-exists'

import { dirname, join, basename } from 'path'

import mapValues from 'lodash/mapValues'
import defaults from 'lodash/defaultsDeep'
import get from 'lodash/get'
import set from 'lodash/set'
import result from 'lodash/result'
import partialRight from 'lodash/partialRight'

import tcomb from 'tcomb-form'

let { isArray } = util

const lodashifyObject = util.lodashify(
   'pick', 'pickBy', 'omit', 'omitBy', 'keys', 'values', 'get', 'set',
   'defaults', 'clone', 'result', 'has', 'hasIn', 'mapValues', 'mapKeys',
   'merge', 'mergeWith', 'invert', 'functions'
)

export class Project {
  static get framework() {
    return require('./index')
  }
  /**
  * Load a Skypager Project from a folder. defaults to process.env.PWD
  *
  * @param {string}   uri                 absolute path to a folder or skypager.js file
  * @param {Object}   options             options for loading the project
  * @param {string}   options.type        an optional type identifier for the project
  * @param {Array}    options.plugins     an array of plugin names, or functions to enable
  * @param {Object}   options.manifest    the package.json manifest data
  * @param {Object}   options.autoLoad    an object specifying which content collections should be autoImported
  * @param {Object}   options.hooks       an object with functions for lifecycle hooks
  * @param {Boolean}  options.autoImport  false to disable autoloading altogether
  *
  * @return {Project}
  */
  static load(...args) {
    return this.framework.load(...args)
  }

  /**
  *
  * @private
  *
  * Wrap a project folder.
  *
  * A Project folder SHOULD contain a package.json with a `skypager` property
  * that contains an object.
  *
  * The `skypager` object on the manifest should contain:
  *
  * - main {string} a file that will get required and export the project.
  *
  * - plugins {Array} an array of the names of plugin packages.
  *   will try to use skypager-plugin-*
  *
  * - provides {Array} an optional list tags for what the project contains
  *     (e.g. ui:components, devtools, themes, website)
  *
  */
  constructor (uri, options = {}) {
    invariant(uri, 'uri must be provided')

    this.setupFsWrapper()

    try {
      normalizeOptions(options)
    } catch(error) {
      console.trace()
      console.log(error)
      throw(error)
    }

    let project = this
    let hide = project.hidden.bind(project)
    let lazy = project.lazy.bind(project)
    let fireHook = project.fireHook.bind(project)

    hide('uri', uri)
    hide('type', options.type)

    project.root = uri.match(/\.(js|json)$/)
      ? dirname(uri)
      : uri

    project.name = util.parameterize(options.name || basename(project.root))
    project.env = util.parameterize(options.env)

    lazy('gitInfo', () => require('./git-info')(this.root))

    hide('createStamp', Math.floor(Date.now() / 100))
    hide('registryId', `${ this.name }:${ this.createStamp }`)

    hide('options', () => options)
    hide('manifest', options.manifest || {})

    // autobind hooks functions passed in as options
    hide('hooks', setupHooks.call(project, options.hooks))
    hide('paths', paths.bind(project))

    logger(project, options)

    hide('schemas', {})
    hide('content', content.call(project))

    hide('registries', registries.call(project), false)

    hide('_run', buildRunInterface.call(project), false)

    if (options.autoImport !== false && options.autoLoad !== false) {
      runImporter.call(project, options.importer)
    }

    project.content.helpers.registryLoaders.map(script => script.register())

    hide('enabledPlugins', [])

    if (options.plugins) {
      options.plugins.forEach(plugin => {
        if (typeof(plugin) === 'function') {
          plugin.call(this, this)
        } else {
          this.use(plugin)
        }
      })
    }

    lazy('entities', () => {
      fireHook('willBuildEntities')
      project.entities = entities.call(project)
      fireHook('didBuildEntities', project, project.entities)

      return project.entities
    })

    if (options.collections || this.setting('collections')) {
      assign(
        project.content,
        buildCustomCollections(this, this.setting('collections'))
      )
    }

    hide('modelDefinitions', modelDefinitions.bind(this))

    hide('framework', () => Project.framework)

    hide('related', {})

    defineProperty(this.related, 'keys', {
      enumerable: false,
      configurable: false,
      value: function(projectName, ...args) {
        return Object.keys(project.related)
      }
    })

    defineProperty(this.related, 'load', {
      enumerable: false,
      configurable: false,
      value: function(projectName, ...args) {
        return project.related[projectName] = project.related[projectName] || project.framework.load(...args)
      }
    })

  }

  get vm () {
    return new VirtualMachine({}, {skypager: this.framework, support: global.$skypager, project: this})
  }

  buildContext(){
    let skypager = project.framework

    if (global.$skypager) {
      let { devpack, ui, electron, server } = global.$skypager

      defineProperty(skypager, 'devpack', {enumerable: false, configurable: true, value: devpack})
      defineProperty(skypager, 'electron', {enumerable: false, configurable: true, value: electron})
      defineProperty(skypager, 'ui', {enumerable: false, configurable: true, value: ui})
      defineProperty(skypager, 'server', {enumerable: false, configurable: true, value: server})
    }

    return new VirtualMachine({},{
      skypager,
      project: this
    })
  }

  lookupGitRemote(remote = 'origin') {
    let info = require('child_process')
      .execSync('git remote show ' + remote).toString().trim()
      .split("\n").filter(line => line.match(/Fetch URL:/))

      if (info.length > 0) {
         return info.shift().trim().split(': ').pop()
      }
  }

  get frameworkPackages () {
    return {
      get ui() { return require(global.$skypager.ui + '/plugin') },
      get electron() { return require(global.$skypager.electron + '/plugin') },
      get devpack() { return require(global.$skypager.devpack + '/plugin') },
      get server() { return require(global.$skypager.server + '/plugin') }
    }
  }

  get helpersRoot () {
    return this.options.helpersRoot
      ? join(this.root, this.options.helpersRoot)
      : this.root
  }

  /**
   * If this project is a portfolio, then it will return itself.
   * If you want to be specific you can say this.parentPortfolio
  */
  get portfolio () {
    if (this.enabledPlugins.indexOf('portfolio') >= 0) {
      return this
    } else {
      return this.parentPortfolio
    }
  }

  get parentPortfolio () {
    if(this.pathExists('../../package.json')) {
      let manifest = require(join(this.root, '../../package.json'))

      if (get(manifest, 'skypager.plugins', []).indexOf('portfolio') >= 0) {
        return this.framework.load(
          join(this.root, '../..')
        )
      }
    }
  }

  /**
   * Returns the skypager project for the ./library
  */
  get library () {
    if (this.pathExists('library')) {
      this.related.load('library', this.paths.join('library'))
      this.related.library.allAssets.forEach(asset => asset.readSync() && asset.assetWasImported())
      return this.related.library
    }
  }

  get fullyLoaded() {
     this.allAssets.forEach(a => a.readSync().assetWasImported())
     return this
  }

  /**
  * Combine all of the project settings files into a single structure.
  *
  * The values for a projects settings will be specific to process.env.NODE_ENV
  * if the settings file contains keys that match development, production, test etc.
  *
  * @return {Object}
  */
  get settings () {
    return this.content.settings_files.notSecret.condense({
      key: 'idpath',
      prop: 'data'
    })
  }

  setting(...args) {
    return this.getSetting(...args)
  }

  getCopy(path, defaultValue) {
    return this.fetchGet(path, this.copy, defaultValue)
  }

  getSetting(path, defaultValue) {
    return this.fetchGet(path, this.settings, defaultValue)
  }

  updateSetting(path, value) {
    if(! this.getSetting(path) ) return false

    let settings_file = this.content.settings_files.resolvePath(path)
    let settingsPath = path.replace(
      `${ settings_file.id.replace(/\//g, '.') }`, ''
    )

    if (settings_file) settings_file.set(settingsPath, value)

    return this
  }

  schema(name, definition) {
    if( !definition ) {  return this.schemas[name] }

    let t = tcomb

    if(!this.schemas[name] && definition) {
      definition = mapValues(
        definition,
        (v) => typeof v === 'string' && this.t[v] ? this.t[v] : v
      )

      let schema = this.schemas[name] = t.struct(definition)

      defineProperty(schema, 'validate', {
        enumerable: false,
        value: partialRight(tcomb.validate, schema)
      })

      return schema
    }

    return this.schemas[name]
  }

  /**
   * Like lodash#result however the path can be an array
   * which will let you try multiple paths before arriving
   * at an optional fallback value.
   *
   * @param {!string} path a path or list of paths which have desired value
   * @param {?Object} resultRoot where to result the value from
   * @param {?whatever} fallbackValue
   *
   * @return {Object} an object with the requested paths set with matching values.
  */
  fetchResult(path, resultRoot=this, fallbackValue) {
    if (isArray(path)) {
      let match = path.find(v => result(resultRoot, v))
      return match ? result(resultRoot, match) : fallbackValue
    }

    return result(resultRoot, path, fallbackValue)
  }

  /**
   * Like lodash#get however the path can be an array
   * which will let you try multiple paths before arriving
   * at an optional fallback value.
   *
   * @param {!string} path a path or list of paths which have desired value
   * @param {?Object} resultRoot where to result the value from
   * @param {?whatever} fallbackValue
   *
   * @return {Object} an object with the requested paths set with matching values.
  */
  fetchGet(path, resultRoot=this, fallbackValue) {
    if (isArray(path)) {
      let match = path.find(v => get(resultRoot, v))
      return match ? get(resultRoot, match) : fallbackValue
    }

    return get(resultRoot, path, fallbackValue)
  }

  /**
   * Create an object from a list of object paths.
   *
   * This uses lodash#pick under the hood. The properties
   * are extracted using lodash#get instead, which lets you
   * pick values nested deeply in the object.
   *
   * @param {...string} paths a list of object paths to pick
   *
   */
  pick(...args) {
    return args.reduce((memo,path) => {
      let value = this.fetchGet(path)
      set(memo, path, value)
      return memo
    }, {})
  }

  /**
   * Create an object from a list of object paths.
   *
   * This uses lodash#pick under the hood. The properties
   * are extracted using lodash#result instead, which lets you
   * pick values nested deeply in the object, executing functions
   * whenever encountering them
   *
   * @param {...string} paths a list of object paths to pick
   *
  */
  pickResult(...args) {
    return args.reduce((memo,path) => {
      let value = this.fetchResult(path)
      set(memo, path, value)
      return memo
    }, {})
  }

  /**
   * Create an object using {Project#pick} and start a promise chain with it.
   *
   * @param {...string} paths dot paths to pick as values
   * @return {Promise<Object>}
   */
  pickPromise(...args) {
    return Promise.resolve(
      this.pick(...args)
    )
  }

  /**
   * Create an object using {Project#pickResult} and start a promise chain with it.
   *
   * @param {...string} paths dot paths to pick as values
   * @return {Promise<Object>}
   *
   */
  pickResultPromise(...args) {
    return Promise.resolve(
      this.pickResult(...args)
    )
  }

  /**
  * Combine all of the project copy files into a single structure.
  *
  * Copy values will be specific to the current locale setting for a project.
  *
  * @type {Object}
  */
  get copy () {
    return this.content.copy_files.query((s) => true).condense({
      key: 'idpath',
      prop: 'data'
    })
  }

  /**
  * Returns the current locale for the project. This value gets used
  * when building the copy system
  *
  * @type {string}
  */
  get locale() {
    return this.getSetting(['app.locale','app.language','locale'], 'en')
  }

  /**
  * Trigger a project lifecycle hook.
  *
  * @param {string} name the name of the life cycle hook
  */
  fireHook(name, ...args) {
    let project = this
    let fn = project.hooks[name] || project[name]
    if (typeof fn === 'function') { fn.call(project, ...args) }
  }

  /**
  * Get an arbitrary value from the project using the lodash result utility.
  *
  * @example
  *
  *   project.get('docs.all[0].paths.absolute')
  */
  get (path, defaultValue) {
    return this.fetchResult(path, this, defaultValue)
  }

  /**
  * Returns an instance of the project's vaule. This is a data store which
  * can be used to store keys and credentials that should not be exposed to the
  * outside, via an exporter, when publishing the project, etc.
  */
  get vault() {
    return vault(this)
  }

  /**
  * Provides a nicer language like interface around looking up and running
  * one of the project's helpers (e.g. actions, exporters, importers, etc.)
  *
  * @example
  *
  *   project.run.importer('disk')
  *   project.run.action('snapshots/save', '/path/to/snapshot.json')
  *
  */
  get run() {
    return this._run
  }

  /**
  * Find a single object in the project using the query mechanism.
  * Accepts any argument that project.query accepts.
  *
  * @param {string} source where to start the search.
  *   Any content collection or model name is valid.
  *
  * @param {Object, Function} params either attributes to match, or a predicate function
  */
  findBy(source, params) {
    params.limit = 1
    return this.query(source, params)[0]
  }

  get querySources() {
    return keys(this.content)
      .concat(
        keys(this.entities || {})
      ).concat([
        'docs', 'data', 'datasources'
      ])
  }

  query (source, params) {
    source = `${ source }`.toLowerCase()

    invariant(
       this.querySources.indexOf(source) >= 0,
       'Must supply a valid source to query from: ' + this.querySources.join(', ')
    )

    if (source === 'docs' || source === 'documents') {
      return this.docs.query(params)
    }

    if (source === 'data' || source === 'datasources' || source === 'data_sources') {
      return this.content.data_sources.query(params)
    }

    if (this.content[source]) {
      return this.content[source].query(params)
    }

    if (this.modelGroups.indexOf(source) > 0) {
      return util.filterQuery(
        util.values(this.entities[source]),
        params
      )
    }
  }

  /**
  * Run a query against the project helper registries.
  *
  * @param {string} source which registry to run your query against
  * @param {Object, Function} params the search criteria
  *
  * @return {Helper}
  */
  queryHelpers(source, params) {
     return this.registries[source].query(params)
  }

  /**
  * Returns a manifest of all of the project's assets.
  */
  get assetManifest () {
    return this.exporters.run('assets', {
      project: this
    })
  }

  /**
   * Returns an array of all of this project's content collections.
   */
  get collections() {
    return util.values(this.content)
  }

  /**
   * lodash mapValues over each of the content collections
  */
  mapCollections(...args) {
    return lodashifyObject(mapValues(this.content, ...args))
  }

  get contentInterface() {
    return this.mapCollections(c => {
      return lodashifyObject(
        c.mapValues(a => lodashifyObject(a))
      )
    })
  }

  lodashifyObject(object) { return lodashifyObject(object) }

  /**
  * Returns an array of every asset in the project
  *
  * @return {Array}
  */
  get allAssets () {
    return util.flatten(this.collections.map(c => c.all))
  }

  /**
  * Returns an array of relative paths for every asset in the project.
  *
  * @return {Array}
  */
  get assetPaths (){
    return this.allAssets.map(a => a.paths.project)
  }

  /**
  * Iterate over every asset in the project.
  */
  eachAsset (...args) {
     return this.allAssets.forEach(...args)
  }

  /**
  * Run a reducer function against every asset in the project.
  *
  */
  reduceAssets(...args) {
     return this.allAssets.reduce(...args)
  }

  /**
  * Convenience method for running Array.map for every asset.
  */
  mapAssets(...args) {
     return this.allAssets.map(...args)
  }

  /**
  * Convenience method for running Array.filter for every asset in the project.
  */
  filterAssets(...args) {
     return this.allAssets.filter(...args)
  }

  /**
  * Access a document by the document id short hand
  *
  */
   at (documentId) {
     return this.documents.at(documentId)
   }

  /**
  * This is a system for resolving paths in the project tree to the
  * appropriate helper, or resolving paths to the links to these paths
  * in some other system (like a web site)
  */
  get resolve () {
    return this.resolver
  }

  /**
  * @alias Project#resolve
  */
  get resolver () {
    return resolver.call(this)
  }

  /**
  * Use a plugin from the plugins registry
  *
  */
  use (plugins, options = {}) {
    if (typeof plugins === 'string') {
      plugins = [plugins]
    }

    plugins
      .map(plugin => this.plugins.lookup(plugin, false))
      .filter(plugin => plugin && plugin.api)
      .forEach(pluginConfig => {
        if (pluginConfig.api.modify) {
          options.project = options.project || this
          pluginConfig.api.modify(options)
        } else if (typeof pluginConfig.api === 'function') {
          pluginConfig.api.call(this, this, pluginConfig)
        }

    })

    this.enabledPlugins.push(...plugins)
  }

  /*
  * Aliases to create hidden and lazy getters on the project
  */
  hidden (...args) { return util.hidden.getter(this, ...args) }
  lazy (...args) { return util.lazy(this, ...args) }

  /**
   * build a path from a base (e.g. documents, models, build)
   * using path.join
   */
  path (base, ...rest) {
    return join(this.paths[base], ...rest)
  }

  join(...args) {
    return this.paths.join(...args)
  }

  /**
  * Collection Accessor Methods
  *
  * These can be used to access document collections within the project
  */
  get docs () {
    return this.documents
  }

  get documents () {
    return this.content.documents
  }

  get data_sources () {
    return this.content.data_sources
  }

  get data () {
    return this.content.data_sources
  }

  get helpers () {
    return this.content.helpers
  }

  get images () {
    return this.content.images
  }

  get packages () {
    return this.content.packages
  }

  get projects () {
    return this.content.projects
  }

  get scripts () {
    return this.content.scripts
  }

  get stylesheets () {
    return this.content.stylesheets
  }

  get stylesheets () {
    return this.content.stylesheets
  }

  get vectors () {
    return this.content.vectors
  }

  get actions () {
    return this.registries.actions
  }

  get contexts () {
    return this.registries.contexts
  }

  get exporters () {
    return this.registries.exporters
  }

  get importers () {
    return this.registries.importers
  }

  get plugins () {
    return this.registries.plugins
  }

  get models () {
    return this.registries.models
  }

  get stores () {
    return this.registries.stores
  }

  get renderers () {
    return this.registries.renderers
  }

  get views () {
    return this.registries.views
  }

  get modelGroups () {
    return this.models.all.map(model =>
      util.tableize(util.underscore(model.name))
    )
  }

  pathExists(...args) {
    return pathExists.sync(
      this.paths.join(...args)
    )
  }

  exists(...args) {
    //this.warn('exists will be deprecated in favor of a more standard api')

    try {
      return pathExists.sync(
        this.path(...args)
      )
    } catch(error) {
       return false
    }
  }

  ensureFolder(...pathArgs) {
    let path = this.path(...pathArgs)
    return this.ensurePath(path)
  }

  ensurePath(...args) {
    let name = args[0]
    if(this.paths[name]) { return this.ensureDirSync(this.path(...args)) }
  }

  /**
   * Status Snapshot
   *
   * @todo
   *
   * for the electron apps I needed a simple way of communicating things between multiple processes
   * such as whether a particular server process is running in the context of
   * a given project and if so on what port.
   *
   */
  get statusSnapshot() {
    try {
      return this.readJsonSync(this.statusSnapshotPath)
    } catch(error) {
      return this.updateStatusSnapshot(false)
    }
  }

  get statusSnapshotPath() {
    return this.path('data_sources', 'status-snapshot-' + this.env + '.json')
  }

  clearStatusSnapshot() {
    return this.updateStatusSnapshot(false)
  }

  updateStatusSnapshot(updates = {}) {
    let newStatus = {}

    if( updates === false ) {
       newStatus = {
         registryId: this.registryId,
         root: this.root
       }
    } else {
      newStatus = this.statusSnapshot || newStatus
      keys(updates).forEach(path => set(newStatus, path, updates[path]))
    }

    this.ensureFolder('data_sources')
    this.outputJsonSync( this.statusSnapshotPath, newStatus, 'utf8')

    return newStatus
  }

  setupFsWrapper(lib = 'fs-promise', methods) {
    require('fs-promise').ensureDirSync = require('fs-promise').mkdirsSync

    wrapFsLib(
      this,
      require(lib),
      methods || ['readFile', 'readJson', 'outputFile', 'outputJson', 'ensureDir', 'ensureFile', 'move', 'remove', 'mkdirs', 'emptyDir']
    )

    this.cp = require('fs-promise').copy
    this.copySync = require('fs-promise').copySync
  }

  get t() {
    return tcomb
  }
}

export default Project

function paths () {
  let project = this

  let conventional = {
    assets: join(this.root, 'assets'),
    actions: join(this.helpersRoot, 'actions'),
    data_sources: join(this.root, 'data'),
    data_files: join(this.root, 'db'),
    documents: join(this.root, 'docs'),
    exporters: join(this.helpersRoot, 'exporters'),
    importers: join(this.helpersRoot, 'importers'),
    models: join(this.helpersRoot, 'models'),
    packages: join(this.root, 'packages'),
    plugins: join(this.helpersRoot, 'plugins'),
    projects: join(this.root, 'projects'),
    renderers: join(this.helpersRoot, 'renderers'),
    vectors: join(this.root, 'assets'),
    images: join(this.root, 'assets'),
    scripts: join(this.root, 'src'),
    stylesheets: join(this.root, 'src'),
    manifest: join(this.root, 'package.json'),
    tmpdir: join(this.root, 'tmp'),
    temp: join(this.root, 'tmp'),
    cache: join(this.root, 'tmp', 'cache'),
    logs: join(this.root, 'log'),
    build: join(this.root, 'dist'),
    public: join(this.root, 'public'),
    settings: join(this.root, 'settings'),
    copy: join(this.root, 'copy'),
    join: function (...args) {
      return join(project.root, ...args)
    }
  }

  let custom = mapValues(
    project.options.paths || project.manifest.paths || {},
    (p) => p.match(/^\//) ? p : join(this.root, p)
  )

  return assign(conventional, custom)
}

function content () { return buildContentCollectionsManually.call(this) }

function runImporter (options = {}) {
  let project = this
  let { autoLoad, type } = options

  project.logger.profile('import starting')

  let result = project.importers.run(type || 'disk', {
    project: this,
    collections: this.content,
    autoLoad
  })

  project.logger.profile('import finishing')

  return result
}

function buildContentCollectionsManually () {
  const project = this
  const paths = project.paths

  let {
    Asset, DataSource, Document,
    CopyFile, Image, HelperScript, Script, Stylesheet,
    ProjectManifest, SettingsFile, Vector
  } = Assets

  return {
    assets: Asset.createCollection(this, {root: this.paths.assets}),
    data_sources: DataSource.createCollection(this, {root: this.paths.data_sources}),
    documents: Document.createCollection(this, {root: this.paths.documents}),
    images: Image.createCollection(this, {root: this.paths.images}),
    scripts: Script.createCollection(this, {root: this.paths.scripts}),
    stylesheets: Stylesheet.createCollection(this, {root: this.paths.stylesheets}),
    vectors: Vector.createCollection(this, {root: this.paths.vectors}),

    /*eslint-disable*/
    helpers: new Collection({
      root: this.helpersRoot,
      project: this,
      assetClass: HelperScript,
      include: '{panels,workspaces,templates,menus,components,services,integrations,clients,actions,importers,exporters,models,plugins,renderers,schemas}/**/*.js'
    }),
    /*eslint-enable*/

    packages: new Collection({
      root: this.paths.packages,
      project: this,
      assetClass: ProjectManifest,
      include: '*/package.json',
      exclude: '**/node_modules'
    }),

    projects: new Collection({
      root: this.paths.projects,
      project: this,
      assetClass: ProjectManifest,
      include: '*/package.json',
      exclude: '**/node_modules'
    }),

    copy_files: new Collection({
      root: this.paths.copy,
      project: this,
      assetClass: CopyFile
    }),

    data_files: new Collection({
      root: this.paths.data_files,
      project: this,
      assetClass: DataFile
    }),

    settings_files: new Collection({
      root: this.paths.settings,
      project: this,
      assetClass: SettingsFile
    })
  }
}

function stores () {
  let project = this
}

function registries () {
  let project = this
  let root = project.root

  let registries = Registry.buildAll(project, Helpers, {root})

  project.fireHook('registriesDidLoad', registries)

  return registries
}

function modelDefinitions() {
  return this.models.available.reduce((memo,id) => {
    let model = this.models.lookup(id)

    assign(memo, {
      get [util.tableize(util.underscore(model.name))](){
        return model.definition
      }
    })

    return memo
  }, {})
}

function entities() {
  let base = this.models.available.reduce((memo,id) => {
    let model = this.models.lookup(id)
    let entities = model.entities = model.entities || {}

    assign(memo, {
      get [util.tableize(util.underscore(model.name))](){
        return entities
      }
    })

    return memo
  }, {})

  let e = {}

  // temp
  Object.keys(base).forEach(key => {
     let values = base[key]
     e[ key.split('/').pop() ] = values
  })

  return e
}

function setupHooks(hooks = {}) {
  let project = this

  return keys(hooks).reduce((memo, hook) => {
    let fn = hooks[hook]

    if (typeof fn === 'function') {
      memo[hook] = fn.bind(project)
    }

    return memo
  }, {})
}

function normalizeOptions (options = {}) {
  if (options.manifest && options.manifest.skypager) {
    options = assign(options, options.manifest.skypager)
  }

  return defaults(options, DefaultOptions)
}

function buildCustomCollections(project, config) {
  return mapValues(project.settings.collections, (cfg, name) => {
    let assetClass = Assets[cfg.assetClass]

    invariant(
      assetClass,
       `Invalid assetClass in custom collections: ${ cfg.assetClass }
        Pick one of: ${  keys(Assets).join(', ') }`.trim()
    )

    invariant(
      cfg.root,
      'Must specify a valid root path for this collection'
    )

    return new Collection({
      root: cfg.root,
      project,
      assetClass,
      name,
      exclude: cfg.exclude,
      include: cfg.include,
      autoLoad: cfg.autoLoad !== false
    })
  })
}

function buildRunInterface() {
    let project = this

    return {
      action (helperId, ...args) {
        return project.actions.run(helperId, ...args)
      },

      importer (helperId, ...args) {
        return project.importers.run(helperId, ...args)
      },

      exporter (helperId, ...args) {
        return project.exporters.run(helperId, ...args)
      },

      model (helperId, ...args) {
        return project.models.run(helperId, ...args)
      },

      plugin (helperId, ...args) {
        return project.plugins.run(helperId, ...args)
      },

      renderer (helperId, ...args) {
        return project.renderers.run(helperId, ...args)
      },
    }

}

const HOOKS = [
  'contentWillInitialize',
  'contentDidInitialize',
  'projectWillAutoImport',
  'projectDidAutoImport',
  'willBuildEntities',
  'didBuildEntities',
  'registriesDidLoad'
]

const DefaultOptions = {
  type: 'project',
  env: process.env.NODE_ENV || 'development',
  importer: {
    type: 'disk',
    autoLoad: {
      documents: true,
      data_sources: (!!process.env.NODE_ENV==='test'),
      settings_files: true,
      copy_files: true,
      scripts: true,
      helpers: true,
      packages: true
    }
  }
}

const { assign, defineProperty, keys } = Object

function wrapFsLib(project, lib, methods) {
  let defineMethods = methods.concat( methods.map(m => m + 'Sync') )

  defineMethods.flatten().forEach((m => {
    defineProperty(project, m, {
      enumerable: false,
      configurable: true,
      value: function(fileOrFolder='', ...args) {
        let abs = fileOrFolder.startsWith(require('path').sep)
            ? fileOrFolder
            : project.paths.join(fileOrFolder)

            try {
              return lib[m] && lib[m](abs, ...args)
            } catch(error) {
            }
      }
    })
  }))

  return project
}
