import { join } from 'path'

import assign from 'object-assign'
import utile from 'utile'

import lodash from 'lodash'

const { defineProperty } = Object

const { set, get, template, templateSettings, kebabCase,
  some, defaults, defaultsDeep, result, pick, has, hasIn, cloneDeep, clone } = lodash

const { isEqual, eq, isMatch, lt, lte, gt, gte, isEmpty, isObject, isString, isNumber, castArray } = lodash

const inflections = utile.inflect
const DOMAIN_REGEX = /^[a-zA-Z0-9_-]+\.[.a-zA-Z0-9_-]+$/

export const dotpath = {
   set, get
}

const equals = eq
const not_equals = (...args) => !eq(...args)
const neq = (...args) => !eq(...args)

export const operators = {
   lt, gt, gte, lte, eq, neq, equals, not_equals
}

module.exports.assign = assign
module.exports.defaults = defaultsDeep
module.exports.pick = pick
module.exports.some = some
module.exports.any = some
module.exports.result = result
module.exports.clone = cloneDeep
module.exports.template = template
module.exports.isObject = isObject
module.exports.isNumber = isNumber
module.exports.isString = isString
module.exports.lodash = lodash

let hidden = {
  getter: function (target, name, fn, configurable = false) {
    defineProperty(target, name, {
      configurable: configurable,
      enumerable: false,
      get: function () {
        return (typeof (fn) === 'function') ? fn.call(target) : fn
      }
    })

		return target
  },

  property: function (target, name, value, configurable = false) {
    defineProperty(target, name, {
      configurable: configurable,
      enumerable: false,
      value: value
    })

		return target
  }
}


module.exports.hidden = hidden
module.exports.hide = hidden

export function humanize (s) {
  return inflections.humanize(s).replace(/-|_/g, ' ')
}

export function titleize (s) {
  return inflections.titleize(humanize(s))
}

export function classify (s) {
  return inflections.classify(s)
}

export function tableize (s) {
  return inflections.tableize(s)
}

export function underscore (s) {
  return tableize(s)
}

export function parameterize (s) {
  s = s.replace(/\\|\//g, '-', '')
  s = s.replace(/[^-\w\s]/g, '')  // remove unneeded chars
  s = s.replace(/^\s+|\s+$/g, '') // trim leading/trailing spaces
  s = s.replace(/[-\s]+/g, '-')   // convert spaces to hyphens
  s = s.toLowerCase()             // convert to lowercase
  return s
}

export function slugify (s) {
  return parameterize(s)
}

export function singularize (word) {
  return inflections.singularize(word)
}

export function pluralize (word) {
  return inflections.pluralize(word)
}

// thanks cheever
export function promiseHash(obj) {
  var awaitables = [];

  var keys = Object.keys(obj);
  for (var i = 0; i < keys.length; i++) {
    var key = keys[i];
    var a$ = obj[key];
    awaitables.push(a$);
  }
  return Promise.all(awaitables).then(function (results) {
    var byName = {};
    for (var i = 0; i < keys.length; i++) {
      var key = keys[i];
      byName[key] = results[i];
    }
    return byName;
  });
}

export function execOutputAsync(command, cwd = process.env.PWD, options = {}) {
	const exec = require('child_process').exec

	return new Promise((resolve, reject) => {
    exec(command, {cwd, ...options}, function (err, stdout, stderr) {
      if (err) {
        reject(err);
      } else {
        resolve(stdout.trim());
      }
    })
	})
}

export function lazy (target, attribute, fn, enumerable = false) {
  defineProperty(target, attribute, {
    configurable: true,
    enumerable: enumerable,
    get: function () {
      delete (target[attribute])

      if (enumerable) {
        return target[attribute] = fn.call(target)
      } else {
        let value = fn.call(target)

        defineProperty(target, attribute, {
          enumerable: false,
          configurable: true,
          value
        })

        return value
      }
    }
  })
}

export function flatten (array) {
  return array.reduce((a, b) => a.concat(b), [])
}

export function singularize (string) {
  return inflections.singularize(string)
}

export function createDelegators (target, source, options = {}) {
  let excludeKeys = options.exclude || options.except || []
  let sourceKeys = Object.keys(source).filter(key => excludeKeys.indexOf(key) === -1)

  sourceKeys.forEach(key => defineProperty(target, key, {
    get: function () {
      return source[key]
    }
  }))
}

export function values (object) {
  return Object.keys(object).map(key => object[key])
}

export function mixin (target, source) {
  target = target.prototype; source = source.prototype

  Object.getOwnPropertyNames(source).forEach(function (name) {
    if (name !== 'constructor') defineProperty(target, name,
      Object.getOwnPropertyDescriptor(source, name))
  })
}

export function access (object, dotted) {
	return get(object, dotted)
}

export function runWithLocals(fn, locals = {}) {
  return new Promise((resolve, reject) => {
    noConflict(fn, locals, undefined, resolve, reject)()
  })
}

export function noConflict (fn, provider = {}, scope, onSuccess, onError) {
  invariant(fn, 'provide a function')
  invariant(provider, 'provide a provider')

  let safe = { }

  return function (...args) {
    Object.keys(provider).forEach(globalProp => {
      if (global.hasOwnProperty(globalProp)) {
        let descriptor = Object.getOwnPropertyDescriptor(global, globalProp)

        if (descriptor && descriptor.configurable) {
          defineProperty(safe, globalProp, descriptor)
        }
      }

      if(onSuccess){
      }

      defineProperty(global, globalProp, Object.getOwnPropertyDescriptor(provider, globalProp))
    })

    let result
    let err

    try {
      if(onSuccess) {
      }
      result = scope ? fn.call(scope, ...args) : fn(...args)
    } catch (e) {
      err = e
    }

    Object.keys(provider).forEach(remove => delete (global[remove]))

    Object.keys(safe).forEach(restore => {
      delete (global[restore])
      Object.defineProperty(global, restore, Object.getOwnPropertyDescriptor(safe, restore))
    })

    if (err) {
      typeof onError === 'function' && onError(err)
      throw(err)
    }

    typeof onSuccess === 'function' && onSuccess(err)
    return result
  }
}

export function carve (dataPath, resultValue, initialValue = {}) {
  dotpath.set(initialValue, dataPath, resultValue)
  return initialValue
}


export function isDomain(value) { return value.match(DOMAIN_REGEX) }

export function isPromise (obj) {
  return !!obj && (typeof obj === 'object' || typeof obj === 'function') && typeof obj.then === 'function'
}

export function isArray(arg) {
  return Object.prototype.toString.call(arg) === '[object Array]'
}

export function isRegex(val) {
  if ((typeof val === 'undefined' ? 'undefined' : typeof(val)) === 'object' &&
      Object.getPrototypeOf(val).toString() === '/(?:)/') {
    return true
  }

  return false
}

export function filterQuery (nodeList = [], params, negate = false) {
  if ( typeof params === 'function' ) {
    return nodeList.filter(params)
  }

  let items = nodeList || []

  return items.filter(node => {
    let matchesParam = Object.keys(params).every(key => {
      let param = params[key]
      let value = result(node, key)

      if (isRegex(param) && param.test(value)) {
        return true
      }

      if (typeof (param)==='string' && value === param) {
        return true
      }

      if (typeof (param)==='number' && value === param) {
        return true
      }

      // treat normal arrays to search for some exact matches
      if (isArray(param) && (isString(value) || isNumber(value))) {
        return some(param, (val) => val === value)
      }

      if (isObject(param) && ( param.value || param.val) && ( param.op || param.operator )) {
        return testWithOperator(param, value)
      }

      if (isObject(param) && isObject(value)) {
        return isMatch(value, param)
      }
    })

    return negate ? !matchesParam : matchesParam
  })
}

export function abort(message) {
   console.log(message.red || message)
   process.exit(1)
}

export function findPackageSync (packageName, root = process.env.PWD) {
  const findModules = require('find-node-modules')
  var path = require('path')

  let moduleDirectories = findModules(root, {relative:false})

  let directory = moduleDirectories.find((p) => {
    let exists = pathExists(join(p, packageName))
    return exists
  })

  if (!directory) {
    try {
      let resolvedPath = path.dirname(require.resolve(packageName))
      return resolvedPath
    } catch(error) {
      //console.log('Error looking up package', packageName, error.message)
    }
  }

  return directory && path.resolve( path.join(directory, packageName))
}

export function findPackage (packageName, options = {}) {
  return new Promise((resolve, reject) => {
    let moduleDirectories = findModules(process.env.PWD, {relative:false})
    let directory = moduleDirectories.find((p) => {
      let exists = pathExists(join(p, packageName))
      return exists
    })

    if (!directory) {
      try {
        let resolvedPath = path.dirname(require.resolve(packageName))
        resolve(resolvedPath)
        return
      } catch(error) {

      }
    }

    if (!directory) { reject(packageName) }

    let result = path.resolve(path.join(directory, packageName))

    if(result) { resolve(result) }
  })
}

export function splitPath(p = '') {
   return path.resolve(p).split(path.sep)
}

export function skypagerBabel() {
  let presets = findPackageSync('babel-preset-skypager')

  try {
    presets ? require('babel-register')({presets}) : require('babel-register')
  } catch(error) {
    abort('Error loading the babel-register library. Do you have the babel-preset-skypager package?')
  }
}


export function pathExists(fp) {
  const fs = require('fs')
	var fn = typeof fs.access === 'function' ? fs.accessSync : fs.statSync

  try {
     fn(fp)
     return true
  } catch(error) {
    return false
  }
}

export function findPackageManifest (cwd = process.env.PWD) {
  return require('findup-sync')('package.json', {cwd})
}

export function loadManifestFromDirectory (cwd = process.env.PWD) {
  return require('findup-sync')('package.json', {cwd})
}

function testWithOperator(param, testValue) {
  let operator = param.operator || param.op
  let checkValue = param.value || param.val
  let testFn = operators[operator] || isEqual

  return testFn(testValue, checkValue)
}

module.exports.lodashify = function lodashify(...methods) {
	return function (object) {
		return methods.reduce((memo,method) => {
			defineProperty(memo, method, {
				configurable: true,
				enumerable: false,
				value: lodash.partial(lodash[method], object)
			})

			return memo
		}, object)
	}
}
